"""Cookieman resources."""

try:
    import importlib.resources as res
except ImportError:  # pragma: no-cover
    import importlib_resources as res


def read_text(resource_name):  # type: (str) -> str
    """
    Get resource content by name.

    :param resource_name: name of resource to load
    :returns: decoded resource content
    """
    return res.read_text(__name__, resource_name)



"""Cookieman flask-compatible session module."""

import six.moves
import flask.wrappers
import flask.sessions
import werkzeug.datastructures

try:
    import typing
except ImportError:
    pass


class LazySession(six.moves.UserDict, flask.sessions.SessionMixin):
    """Session mapping with deferred data loading."""

    callback_dict_class = werkzeug.datastructures.CallbackDict

    modified = False
    accessed = False

    def _notify_update(self, data):  # type: (typing.Any) -> None
        """
        Flag this object as modified.

        Used as werkzeug CallbackDict callback.

        :param data: ignored
        """
        self.modified = True
        self.accessed = True

    def __init__(self,
                 session_data_fnc,  # type: typing.Callable[[], dict]
                 request=None,  # type: typing.Optional[flask.wrappers.Request]
                 ):  # type: (...) -> None
        """
        Initialize.

        :param session_data_fnc: session data getter function
        :param request: request object
        """
        self._data = None
        self._session_data_fnc = session_data_fnc
        self._request = request

    @property
    def data(self):  # type: () -> typing.Mapping
        """
        Return session data mapping.

        :return: inner session mapping object
        """
        self.accessed = True
        if self._data is None:
            self._data = self.callback_dict_class(
                self._session_data_fnc(),
                self._notify_update,
                )
        return self._data

    # TODO (on python2 drop): remove
    def __iter__(self):  # type: () -> typing.Iterator[typing.Any]
        """
        Return iter(x).

        :return: keys in session mapping
        """
        return iter(self.data)



import re

import unittest_resources.testing as base


class Mixin:
    """Package mixin."""

    meta_module = 'cookieman'
    meta_module_pattern = re.compile(r'^([^t]*|t(?!ests?\.))+$')
    meta_resource_pattern = re.compile(r'^([^t]*|t(?!ests?))+\.py$')


class TypingTestCase(Mixin, base.TypingTestCase):
    """TestCase checking :module:`mypy`."""


class CodeStyleTestCase(Mixin, base.CodeStyleTestCase):
    """TestCase checking :module:`pycodestyle`."""


class DocStyleTestCase(Mixin, base.DocStyleTestCase):
    """TestCase checking :module:`pydocstyle`."""


class MaintainabilityIndexTestCase(Mixin, base.MaintainabilityIndexTestCase):
    """TestCase checking :module:`radon` maintainability index."""


class CodeComplexityTestCase(Mixin, base.CodeComplexityTestCase):
    """TestCase checking :module:`radon` code complexity."""

    max_class_complexity = 6
    max_function_complexity = 9



"""Cookieman compressed, paginated and signed session cookies."""

__version__ = '0.0.3'

import datetime

try:
    import typing
except ImportError:
    pass

import flask
import flask.wrappers
import flask.sessions
import flask.helpers

import cookieman.cookie as cookie
import cookieman.serializer as serializer
import cookieman.session as session
import cookieman.limits as limits


class CookieMan(flask.sessions.SessionInterface):
    """Compressed, paginated and signed session cookie manager."""

    limits_class = limits.Limits
    shrink_class = limits.ShrinkManager
    cookie_processor_class = cookie.CookieProcessor
    serializer_class = serializer.CookieManSerializer
    session_class = session.LazySession

    def __init__(self, salt='session-cookie'):  # type: (str) -> None
        """
        Initialize.

        :param salt: cookie signature salt
        """
        self._cookie_processor = self.cookie_processor_class(self)
        self._cookie_shrink = self.shrink_class()
        self._limits_cache = {}
        self.salt = salt

    def register(self,
                 key_or_keys,  # type: typing.Union[str, typing.Iterable[str]]
                 shrink_fnc=None,  # type: limits.ShrinkFunction
                 ):  # type: (...) -> limits.ShrinkFunction
        """
        Register a session shrinking function for given keys.

        This method can be used as decorator.

        :param key_or_keys: key or iterable of keys would be affected
        :param shrink_fnc: shrinking function (optional for decorator)
        :returns: either original given shrink_fnc or decorator

        Usage:

        .. code-block:: python

            @app.session_interface.register('my_session_key')
            def my_shrink_fnc(data):
                del data['my_session_key']
                return data

        """
        return self._cookie_shrink.register(key_or_keys, shrink_fnc)

    def get_browser(self,
                    app,  # type: flask.Flask
                    request,  # type: flask.wrappers.Request
                    ):  # type: (...) -> limits.Browser
        """
        Get browser limits object based on app and request.

        :param app: flask app object
        :returns: limits object
        """
        limits_file = app.config.get('COOKIEMAN_LIMITS_PATH', None)
        if limits_file not in self._limits_cache:
            self._limits_cache[limits_file] = \
                self.limits_class.load(limits_file)
        return self._limits_cache[limits_file].get(
            request.user_agent.browser,
            request.user_agent.platform,
            request.user_agent.version,
            )

    def get_signing_serializer(self,
                               app,  # type: flask.Flask
                               ):
        # type: (...) -> typing.Optional[serializer.CookieManSerializer]
        """
        Get signing serializer class instance for given app.

        :param app: application instance
        :returns: instance if app has secret key else None
        :rtype: None or serializer class
        """
        if not app.secret_key:
            return None
        return self.serializer_class(app.secret_key, self.salt)

    def get_expiration_time(self,
                            app,  # type: flask.Flask
                            session=None,  # type: typing.Any
                            ):  # type: (...) -> datetime.datetime
        """
        Get expiration time for a new cookie based on app config.

        :param app: flask app
        :param session: ignored
        ;returns: expiration time
        """
        return datetime.datetime.utcnow() + app.permanent_session_lifetime

    def open_session(self,
                     app,  # type: flask.Flask
                     request,  # type: flask.wrappers.Request
                     ):  # type: (...) -> typing.Optional[session.LazySession]
        """
        Create lazy session from request cookies.

        :param app: flask app object
        :param request: request object to extract
        :returns: session dict-like object
        """
        def extract_cookie_data():
            data = b''.join(
                cookie.value
                for cookie in self._cookie_processor.iter_request_cookies(
                    app,
                    request,
                    self.get_browser(app, request),
                    )
                )
            if data:
                max_age = flask.helpers.total_seconds(
                    app.permanent_session_lifetime
                    )
                try:
                    return serializer.loads(data, max_age=max_age)
                except Exception:
                    pass
            return {}

        serializer = self.get_signing_serializer(app)
        if serializer:
            return self.session_class(extract_cookie_data, request)
        return None

    def save_session(self,
                     app,  # type: flask.Flask
                     session,  # type: session.LazySession
                     response,  # type: flask.wrappers.Response
                     ):  # type: (...) -> None
        """
        Update response in-place with session cookies.

        :param app: flask app object
        :param session: cookieman session object
        :param response: flask response object
        :returns: given response object (with cookie changes)
        """
        request = getattr(session, '_request', None) or flask.request
        browser = self.get_browser(app, request)
        cookies = self._cookie_processor

        if session.modified and not session:
            options = cookies.get_cookie_options(app, browser)
            response.delete_cookie(
                app.session_cookie_name,
                domain=options['domain'],
                path=options['path'],
                )
            return

        if session.accessed and session:
            response.vary.add('Cookie')

        if not self.should_set_cookie(app, session):
            return

        iter_cookies = cookies.iter_response_cookies
        cookies = self._cookie_shrink.shrink(
            dict(session),
            browser,
            self.get_signing_serializer(app).dumps,
            lambda payload: iter_cookies(app, payload, browser)
            )

        for c in cookies:
            response.set_cookie(c.name, c.value, **c.options)



"""Cookieman serializer submodule."""

import zlib
import base64

try:
    import typing
except ImportError:
    pass

import msgpack
import itsdangerous


class CookieManSerializer(object):
    """
    Signed compressed session cookie serializer.

    This class implements the following:

    - Messagepack python object serialization.
    - Signature check using itsdangerous.TimestampSigner.
    - Compression using python's standard zlib.
    - URL-safe base64 encoding (to http header issues).

    """

    signer_class = itsdangerous.TimestampSigner

    def __init__(self, secret, salt):  # type: (str, str) -> None
        """
        Initialize.

        :param secret: itsdangerous signer secret
        :param salt: itsdangerous signer salt
        """
        self.signer = self.signer_class(secret, salt=salt)

    def dumps(self, data):  # type: (typing.Any) -> bytes
        """
        Get serialized representation of given object.

        :param data: serializable object
        :return: serialized data
        """
        dumped = msgpack.packb(data, use_bin_type=True)
        signed = self.signer.sign(dumped)
        compressed = zlib.compress(signed)
        return base64.urlsafe_b64encode(compressed)

    def loads(self,
              data,  # type: bytes
              max_age=None,  # type: typing.Optional[int]
              ):  # type: (...) -> typing.Any
        """
        Get object from serialized data.

        :param data: serialized data
        :return: unserialized object
        """
        decoded = base64.urlsafe_b64decode(data)
        decompressed = zlib.decompress(decoded)
        signed = self.signer.unsign(decompressed, max_age=max_age)
        return msgpack.unpackb(signed, raw=False)



"""Cookieman browser limits and shrinking classes."""

import io
import sys
import itertools
import functools
import collections

import six
import six.moves
import werkzeug.datastructures

import cookieman.resources as resources
import cookieman.exceptions as exceptions

try:
    import typing
    TCookie = typing.TypeVar('TCookie', bound=typing.Sized)
    LimitsSection = typing.Tuple[str, typing.Mapping[str, str]]
    ShrinkFunction = typing.Callable[
        [typing.MutableMapping, bool],
        typing.MutableMapping
        ]
    ShrinkFunctionGenerator = typing.Generator[
        typing.Callable[
            [typing.MutableMapping],
            typing.MutableMapping
            ],
        typing.Tuple[typing.Iterable[str], bool],
        None
        ]
except ImportError:
    pass


class Browser(object):
    """
    Browser limits.

    :attr name: browser name
    :type name: str
    :attr platform: browser platform (generic OS name)
    :type platform: str
    :attr version: browser name
    :type version: Tuple[Union[int, str]]
    :attr maxcookies: maximum supported number of cookies
    :type maxcookies: int
    :attr maxsize: maximum supported size of a single cookie
    :type maxsize: int
    :attr maxtotal: maximum total supported size of all cookies
    :type maxtotal: int
    """

    __slots__ = (
        'name', 'platform', 'version', 'maxcookies', 'maxsize',
        'maxtotal',
        )

    def __init__(self,
                 name='',  # type: str
                 platform='',  # type: str
                 version=(),  # type: typing.Iterable[str]
                 maxcookies=1,  # type: int
                 maxsize=4096,  # type: int
                 maxtotal=None,  # type: typing.Optional[int]
                 ):  # type: (...) -> None
        """
        Initialize.

        :param name: browser name
        :param platform: browser platform
        :param version: browser version
        :param maxcookies: maximum number of cookies
        :param maxsize: maximum number of bytes of every cookies
        :param maxtotal: maximum number of total bytes of all cookies
        """
        self.name = name
        self.platform = platform
        self.version = tuple(version)
        self.maxcookies = maxcookies
        self.maxsize = maxsize
        self.maxtotal = (
            min(maxtotal, maxcookies * maxsize)
            if maxcookies and maxtotal else
            maxtotal
            if maxtotal else
            maxcookies * maxsize
            )

    def __repr__(self):  # type: () -> str
        """
        Return repr(self).

        :return: string representation of object
        """
        name = '%s.%s' % (
            self.__class__.__module__,
            self.__class__.__name__
            )
        props = {attr: getattr(self, attr) for attr in self.__slots__}
        return '<%s%r>' % (name, props)

    @classmethod
    def parse_version(cls,
                      version,  # type:  typing.Iterable[str]
                      ):
        # type: (...) -> typing.Iterable[typing.Union[int, str]]
        """
        Convert version (str or tuple) to version tuple.

        :param version: version value (str or tuple)
        :type version:
        :return: version tuple
        """
        string_types = six.string_types
        if version:
            if isinstance(version, string_types):
                version = version.split('.')
            return tuple(
                int(i) if isinstance(i, string_types) and i.isdigit() else i
                for i in version
                )
        return ()

    @classmethod
    def from_spec(cls,
                  name,  # type: str
                  properties,  # type: typing.Mapping[str, str]
                  ):  # type: (...) -> Browser
        """
        Create browser object based on limits database data.

        :param name: browser identifier (name:platform:version)
        :param properties:
        :returns: browser object
        """
        name, platform, version = name.split(':')[:3]
        return cls(
            name,
            platform or '',
            cls.parse_version(version),
            int(properties['maxcookies']),
            int(properties['maxsize']),
            int(properties['maxtotal']),
            )


class Limits(object):
    """Browser limits database."""

    browser_class = Browser
    mapping_class = werkzeug.datastructures.ImmutableDict

    def __init__(self,
                 browsers=(),  # type: typing.Iterable[Browser]
                 ):  # type: (...) -> None
        """
        Initialize.

        :param browsers: iterable with all available browsers
        """
        self.browsers_by_key = self.mapping_class(
            (key, tuple(browsers)) for key, browsers in itertools.groupby(
                sorted(
                    browsers,
                    key=lambda browser: (browser.name, browser.platform,
                                         browser.version),
                    reverse=True,
                ),
                key=lambda browser: (browser.name, browser.platform)))
        self.default = self.browser_class()

    def get(self,
            name='',  # type: str
            platform='',  # type: str
            version=(),  # type: typing.Iterable[str]
            ):  # type: (...) -> Browser
        """
        Get browser limits object matching giving criteria.

        :param name: browser name
        :param platform: browser platform name
        :param version: browser version tuple
        :return: browser limits object
        """
        version = self.browser_class.parse_version(version)
        search_keys = (
            ((name, platform), version),
            ((name, platform), ()),
            ((name, ''), version),
            ((name, ''), ()),
            (('', ''), ()),  # database default
            )
        for (key, version) in search_keys:
            for browser in self.browsers_by_key.get(key, ()):
                if browser.version <= version:
                    return browser
        return self.default

    def __eq__(self, value):  # type: (Limits) -> bool
        """Return self==value."""
        return (
            isinstance(value, self.__class__) and
            self.browsers_by_key == value.browsers_by_key
            )

    if six.PY2:
        @classmethod
        def get_sections(cls, data):
            # type: (str) -> typing.Generator[LimitsSection, None, None]
            """Yield sections from config data."""
            parser = six.moves.configparser.SafeConfigParser()
            with io.StringIO(data) as f:
                parser.readfp(f)
                for name in parser.sections():
                    yield name, dict(parser.items(name))
    else:
        @classmethod
        def get_sections(cls, data):
            # type: (str) -> typing.Generator[LimitsSection, None, None]
            """Yield sections from config data."""
            parser = six.moves.configparser.ConfigParser()
            parser.read_string(data)

            for name in parser.sections():
                yield name, parser[name]

    @classmethod
    def load(cls, path=None):  # type: (typing.Optional[str]) -> Limits
        """
        Load browser limits database file.

        If no path is given, cookieman builtin limits database will be used.

        :param path: path or None
        :returns: Limits object
        """
        if path is None:
            data = resources.read_text('limits.ini')
        else:
            with open(path, 'rb') as f:
                data = f.read().decode('utf-8')

        return cls(
            cls.browser_class.from_spec(name, value)
            for name, value in cls.get_sections(data)
            )


class ShrinkManager(object):
    """Cookie shrinking function manager."""

    _key_added_exc = type('KeyAdded', (Exception, ), {})

    def __init__(self):
        """Initialize."""
        self.handlers = collections.defaultdict(set)

    if typing:
        @typing.overload
        def register(self,
                     key_or_keys,  # type: typing.Iterable[str]
                     shrink_fnc=None,  # type: typing.Literal[None]
                     ):
            # type: (...) -> typing.Callable[[ShrinkFunction], ShrinkFunction]
            """Define register typing overload. Implemented below."""

        @typing.overload
        def register(self,
                     key_or_keys,  # type: typing.Iterable[str]
                     shrink_fnc,  # type: ShrinkFunction
                     ):
            # type: (...) -> ShrinkFunction
            """Define register typing overload. Implemented below."""

    def register(self, key_or_keys, shrink_fnc=None):
        """
        Register a session shrinking function for given keys.

        This method can be used as decorator.

        :param key_or_keys: key or list of keys would be affected
        :param shrink_fnc: shrinking function (optional for decorator)
        :returns: either original given shrink_fnc or decorator

        Usage:

        .. code-block:: python

            @shrinker.register('my_session_key')
            def my_shrink_fnc(data):
                del data['my_session_key']
                return data

        """
        if shrink_fnc is None:
            return functools.partial(self.register, key_or_keys)

        keys = (
            (key_or_keys,)
            if isinstance(key_or_keys, six.string_types) else
            key_or_keys
            )
        for name in keys:
            self.handlers[name].add(shrink_fnc)
        return shrink_fnc

    @classmethod
    def _purge(cls,
               key,  # type: str
               data,  # type: typing.MutableMapping
               last,  # type: bool
               ):  # type: (...) -> typing.MutableMapping
        """
        Remove an unhandled key.

        :param key: data dict key to remove
        :param data: data mapping
        :param last: ignored
        :returns: updated mapping
        """
        del data[key]
        return data

    def _iter_handled(self, keys, last=False):
        # type: (typing.Iterable[str], bool) -> ShrinkFunctionGenerator
        """
        Iterate shrinking functions with fixed second (last) parameter.

        :param keys:
        :param last:
        :returns: generator of callables
        """
        again = True
        handlers = self.handlers
        visited = set()
        while again:
            again = False
            for key in tuple(keys):
                if key in visited:
                    continue

                revisit = False
                for handler in handlers.get(key, ()):
                    if key not in keys:
                        revisit = False
                        break

                    keys, shrunk = (yield lambda data: handler(data, last))
                    revisit |= shrunk

                if revisit:
                    again = True
                else:
                    visited.add(key)

    def _iter_purge(self, keys, last=False):
        # type: (typing.Iterable[str], bool) -> ShrinkFunctionGenerator
        """
        Iterate purge functions with fixed key and last parameters.

        :param keys:
        :param last:
        :return: generator of callables
        """
        handlers = self.handlers
        for key in tuple(keys):
            if last or key not in handlers:
                yield (lambda data: self._purge(key, data, last))

    def _iter_tasks(self, keys):
        # type: (typing.Iterable[str]) -> ShrinkFunctionGenerator
        """
        Iterate functions from different shrinking strategies.

        Generators emits functions accepting the mapping and receiving
        both data and wether data size changes.

        :param keys: possible handler keys to iterate
        :return: generator of callables

        Iteration order:

        1. Shrinking functions for keys on data.
        2. Purging functions for unhandled keys.
        3. Shrinking functions for keys on data (receiving last=True)
        4. Purging functions for every key.

        """
        strategies = [
            lambda keys: self._iter_handled(keys),
            lambda keys: self._iter_purge(keys),
            lambda keys: self._iter_handled(keys, True),
            lambda keys: self._iter_purge(keys, True),
            ]
        for strategy in strategies:
            strategy_iter = strategy(keys)
            try:
                keys, changed = (yield next(strategy_iter))
                while True:
                    keys, changed = (yield strategy_iter.send((keys, changed)))
            except StopIteration:
                pass

    def shrink(self,
               data,  # type: typing.MutableMapping
               browser,  # type: Browser
               dump_fnc,  # type: typing.Callable[[typing.Any], bytes]
               serialize_fnc,  # type: typing.Callable[[bytes], TCookie]
               ):  # type: (...) -> typing.Iterable[TCookie]
        """
        Apply shrink functions to data until under browser limits.

        If it's not possible, raises error.

        :param data: mapping
        :param browser: browser limits object
        :param dump_fnc: callable which returns binary data payload
        :param serialize_fnc: callable which returns cookies
        :return: multiple paginates cookies
        :raises exceptions.CookieSizeException: if unable to shrink enough
        """
        handlers_iter = None
        old_size = sys.maxsize
        maxsize = browser.maxsize
        maxcookies = browser.maxcookies

        while True:
            payload = dump_fnc(data)
            cookies = list(serialize_fnc(payload))
            size = sum(map(len, cookies))

            if size <= maxsize and len(cookies) <= maxcookies:
                return cookies

            try:
                if handlers_iter:
                    shrunk = len(payload) < old_size
                    handler = handlers_iter.send((data, shrunk))
                else:
                    handlers_iter = iter(self._iter_tasks(data))
                    handler = next(handlers_iter)
            except StopIteration:
                break

            old_size = len(payload)
            data = handler(data)

        raise exceptions.CookieSizeException(
            'Session cookies exceeded browser maxsize %d' % (
                browser.maxsize,
                )
            if len(cookies) > maxsize else
            'Session cookies exceeded browser maxcookies %d' % (
                browser.maxcookies
                )
            )



"""Cookieman cookie classes."""

import functools

import six
import werkzeug.http
import werkzeug.datastructures
import flask
import flask.wrappers

import cookieman.exceptions as exceptions
import cookieman.limits as limits

try:
    import typing
    PartitionNameFunction = typing.Callable[[str, int], str]
except ImportError:
    pass


@functools.total_ordering
class Cookie(object):
    """Cookieman cookie."""

    header = 'Set-Cookie: '
    mapping_class = werkzeug.datastructures.ImmutableDict
    value_types = six.binary_type if six.PY2 else (six.binary_type, memoryview)
    dump_cookie_fnc = staticmethod(werkzeug.http.dump_cookie)
    byteview_fnc = staticmethod(memoryview.tobytes if six.PY2 else memoryview)

    @classmethod
    def default_partition_name_format(cls, name, index):
        # typing: (str, int) -> str
        """Get default partition name."""
        return '{}-{:x}'.format(name, index) if index else name

    @property
    def name(self):  # typing: () -> str
        """Get cookie name."""
        return self._name

    @property
    def value(self):  # typing: () -> str
        """Get cookie value."""
        return self._value

    @property
    def options(self):  # typing: () -> str
        """Get cookie options."""
        return self._options

    @property
    def data(self):  # type: () -> str
        """Get cookie header data."""
        if self._data is None:
            self._data = self.header + self.dump_cookie_fnc(
                self._name,
                self._value,
                max_size=0,
                **self._options
                )
        return self._data

    def __init__(self,
                 name,  # type: str
                 value=b'',  # type: bytes
                 options=None,  # type: typing.Optional[typing.Mapping]
                 ):  # type: (...) -> None
        """
        Initialize.

        :param name: cookie name
        :param value: cookie value
        :param options: cookie options
        """
        assert \
            isinstance(value, self.value_types), \
            TypeError('Cannot convert %r object to cookie' % type(value))
        self._data = None
        self._name = name
        self._value = value
        self._options = self.mapping_class(options or ())

    def __eq__(self, value):  # type: (Cookie) -> bool
        """Return self==value."""
        return self.data == value.data

    def __lt__(self, value):  # type: (Cookie) -> bool
        """Return self<value."""
        return (len(self.name), self.name) < (len(value.name), value.name)

    def __len__(self):  # type: () -> int
        """Return len(self)."""
        return len(self.data)

    def __repr__(self):  # type: () -> str
        """Return repr(self)."""
        return '<%s %r>' % (self.__class__.__name__, self.data)

    def partition(self,
                  maxsize,  # type: int
                  partition_name_fnc=None,  # type: PartitionNameFunction
                  ):  # type: (...) -> typing.Generator[Cookie, None, None]
        """
        Generate multiple cookies by spliting cookie value in chunks.

        :param maxsize: maximum total cookie size (including header)
        :param fmt_fnc: cookie name formatting
        :return: iterator of cookies
        """
        name_fnc = functools.partial(
            self.default_partition_name_format
            if partition_name_fnc is None else
            partition_name_fnc,
            self.name
            )
        byteview = self.byteview_fnc

        if len(self) < maxsize and name_fnc(0) == self.name:
            yield self
            return

        cls = self.__class__
        total = len(self.value)
        buffer = memoryview(self.value)
        bodysize = maxsize - len(cls('k', b'a', self.options)) + 2

        page = 0
        start = 0
        lastsize = 0

        while start < total or lastsize == maxsize:
            name = name_fnc(page)

            # retry on cookie excess due expansions
            for end in range(start + bodysize - len(name), start, -1):
                cookie = cls(name, byteview(buffer[start:end]), self.options)
                lastsize = len(cookie)

                if lastsize <= maxsize:
                    break
            else:
                raise exceptions.CookieSizeException(
                    'Unable to paginate cookies by size %d' % maxsize
                    )

            yield cookie
            start = end
            page += 1


class CookieProcessor(object):
    """Compressed-paginated and signed cookie manager."""

    cookie_class = Cookie

    def __init__(self,
                 session_interface,  # type: flask.sessions.SessionInterface
                 ):  # type: (...) -> None
        """
        Initialize.

        :param session_interface: flask session interface instance
        """
        self.session_interface = session_interface

    def get_cookie_options(self,
                           app,  # type: flask.Flask
                           browser,  # type: limits.Browser
                           ):  # type: (...) -> typing.Mapping[str, typing.Any]
        """
        Get cookie options dict as accepted by flask cookie dump methods.

        :param app: flask app
        :return: dict with options
        """
        return {
            'expires': self.session_interface.get_expiration_time(app),
            'path': self.session_interface.get_cookie_path(app),
            'domain': self.session_interface.get_cookie_domain(app),
            'secure': self.session_interface.get_cookie_secure(app),
            'httponly': self.session_interface.get_cookie_httponly(app),
            }

    def iter_request_cookies(self,
                             app,  # type: flask.Flask
                             request,  # type: flask.wrappers.Request
                             browser,  # type: limits.Browser
                             ):
        # type: (...) -> typing.Generator[cookie.Cookie, None, None]
        """
        Extract data from cookies.

        :param app: flask app
        :param request: flask request
        :param browser: cookieman browser limits
        :return: iterator of paginated cookies
        """
        cookies = request.cookies
        cookie_name = app.session_cookie_name
        cookie_prefix = '%s-' % app.session_cookie_name

        try:
            cookies = sorted(
                self.cookie_class(key, value.encode('ascii'))
                for key, value in request.cookies.items()
                if key == cookie_name or key.startswith(cookie_prefix)
                )
        except UnicodeEncodeError:
            cookies = ()

        if cookies:
            maxsize = len(cookies[0])
            for cookie in cookies:
                size = len(cookie)

                # spurious cookie
                if size > maxsize:
                    break

                yield cookie

                # paginated cookies have the same size except for the last one
                if size < maxsize:
                    break

    def iter_response_cookies(self,
                              app,  # type: flask.Flask
                              data,  # type: bytes
                              browser,  # type: limits.Browser
                              ):
        # (...) -> typing.Generator[cookie.Cookie, None, None]
        """
        Split given byte string for cookies.

        :param app: flask app
        :param data: byte string
        :param browser: cookieman browser limits
        :return: iterator of paginated cookie
        """
        options = self.get_cookie_options(app, browser)
        cookie = self.cookie_class(app.session_cookie_name, data, options)
        for page in cookie.partition(browser.maxsize):
            yield page
